import json

from spotinst_sdk2.client import Client
import spotinst_sdk2.models.functions as spotinst_functions


class FunctionsClient(Client):
    __base_functions_url = "/functions"

    # region Functions
    def create_application(self, app):
        """
        Create Spotinst Functions Application

        # Arguments
        app (ApplicationCreate): ApplicationCreate Object
        # Returns
        (Object): Functions API response 
        """
        app = spotinst_functions.ApplicationCreationRequest(app)

        excluded_group_dict = self.exclude_missing(json.loads(app.toJSON()))

        formatted_app_dict = self.convert_json(
            excluded_group_dict, self.underscore_to_camel)

        body_json = json.dumps(formatted_app_dict)

        self.print_output(body_json)

        app_response = self.send_post(
            body=body_json,
            url=self.__base_functions_url +
            '/application',
            entity_name='application')

        formatted_response = self.convert_json(
            app_response, self.camel_to_underscore)

        retVal = formatted_response["response"]["items"][0]

        return retVal

    def create_environment(self, env):
        """
        Create Spotinst Functions Environment

        # Arguments
        env (EnvironmentCreate): EnvironmentCreate Object
        # Returns
        (Object): Functions API response 
        """
        env = spotinst_functions.EnvironmentCreationRequest(env)

        excluded_env_dict = self.exclude_missing(json.loads(env.toJSON()))

        formatted_env_dict = self.convert_json(
            excluded_env_dict, self.underscore_to_camel)

        body_json = json.dumps(formatted_env_dict)

        self.print_output(body_json)

        env_response = self.send_post(
            body=body_json,
            url=self.__base_functions_url +
            '/environment',
            entity_name='environment')

        formatted_response = self.convert_json(
            env_response, self.camel_to_underscore)

        retVal = formatted_response["response"]["items"][0]

        return retVal

    def create_function(self, fx):
        """
        Create Spotinst Functions

        # Arguments
        fx (FunctionCreate): FunctionCreate Object
        # Returns
        (Object): Functions API response 
        """
        fx = spotinst_functions.FunctionCreationRequest(
            fx, self.should_print_output)

        excluded_fx_dict = self.exclude_missing(json.loads(fx.toJSON()))

        formatted_fx_dict = self.convert_json(
            excluded_fx_dict, self.underscore_to_camel)

        body_json = json.dumps(formatted_fx_dict)

        formatted_fx_dict['function']['code']['source'] = 'INLINE_BASE64_SOURCE_CODE'
        self.print_output(json.dumps(formatted_fx_dict))

        fx_response = self.send_post(
            body=body_json,
            url=self.__base_functions_url +
            '/function',
            entity_name='function')

        formatted_response = self.convert_json(
            fx_response, self.camel_to_underscore)

        retVal = formatted_response["response"]["items"][0]

        return retVal

    # endregion



import json

none = "d3043820717d74d9a17694c176d39733"

# region EMR


class BlueGreenDeployment:
    """
    # Arguments
    timeout: int
    tags: List[Tag]
    deployment_groups: List[DeploymentGroup]
    """

    def __init__(
            self,
            timeout=none,
            tags=none,
            deployment_groups=none):

        self.timeout = timeout
        self.tags = tags
        self.deployment_groups = deployment_groups

# endregion


class Tag:
    """
    # Arguments
    tag_get: str
    tag_value: str
    """

    def __init__(
            self,
            tag_key=none,
            tag_value=none):

        self.tag_key = tag_key
        self.tag_value = tag_value


class DeploymentGroup:
    """
    # Arguments
    application_name: str
    deployment_group_name: str
    """

    def __init__(
            self,
            application_name=none,
            deployment_group_name=none):

        self.application_name = application_name
        self.deployment_group_name = deployment_group_name


class BlueGreenDeploymentRequest:
    def __init__(self, deployment):
        self.deployment = deployment

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)



import json
from enum import Enum
from typing import List

none = "d3043820717d74d9a17694c176d39733"

# region Verification Arguments


class SecretKeyRef:
    """
    # Arguments
    key : str
    name : str
    """

    def __init__(
            self,
            key: str = none,
            name: str = none):
        self.key = key
        self.name = name


class VerificationTemplateArgument:
    """
    # Arguments
    name : str
    secret_key_ref : SecretKeyRef
    value : str
    """

    def __init__(
            self,
            name: str = none,
            secret_key_ref: SecretKeyRef = none,
            value: str = none):
        self.name = name
        self.secret_key_ref = secret_key_ref
        self.value = value
# endregion


# region Verification Metrics
class Prometheus:
    """
    # Arguments
    query : str
    """

    def __init__(
            self,
            query: str = none):
        self.query = query


class NewRelic:
    """
    # Arguments
    profile : str
    query : str
    """

    def __init__(
            self,
            profile: str = none,
            query: str = none):
        self.profile = profile
        self.query = query


class Datadog:
    """
    # Arguments
    duration : str
    query : str
    """

    def __init__(
            self,
            duration: str = none,
            query: str = none):
        self.duration = duration
        self.query = query


class Dimensions:
    """
    # Arguments
    name : str
    value : str
    """

    def __init__(
            self,
            name: str = none,
            value: str = none):
        self.name = name
        self.value = value


class MetricUnit(Enum):
    Seconds = "Seconds"
    Microseconds = "Microseconds"
    Milliseconds = "Milliseconds"
    Bytes = "Bytes"
    Kilobytes = "Kilobytes"
    Megabytes = "Megabytes"
    Gigabytes = "Gigabytes"
    Terabytes = "Terabytes"
    Bits = "Bits"
    Kilobits = "Kilobits"
    Megabits = "Megabits"
    Gigabits = "Gigabits"
    Terabits = "Terabits"
    Percent = "Percent"
    Count = "Count"
    BytesBySecond = "Bytes/Second"
    KilobytesBySecond = "Kilobytes/Second"
    MegabytesBySecond = "Megabytes/Second"
    GigabytesBySecond = "Gigabytes/Second"
    TerabytesBySecond = "Terabytes/Second"
    BitsBySecond = "Bits/Second"
    KilobitsBySecond = "Kilobits/Second"
    MegabitsBySecond = "Megabits/Second"
    GigabitsBySecond = "Gigabits/Second"
    TerabitsBySecond = "Terabits/Second"
    CountBySecond = "Count/Second"
    NONE = "None"


class Metric:
    """
    # Arguments
    metric_name : str
    namespace : str
    dimensions : List[Dimensions]
    """

    def __init__(
            self,
            metric_name: str = none,
            namespace: str = none,
            dimensions: List[Dimensions] = none):
        self.metric_name = metric_name
        self.namespace = namespace
        self.dimensions = dimensions


class MetricStat:
    """
    # Arguments
    metric : Metric
    period : int
    stat : str
    unit : MetricUnit
    """

    def __init__(
            self,
            metric: Metric = none,
            period: int = none,
            stat: str = none,
            unit: MetricUnit = none):
        self.metric = metric
        self.period = period
        self.stat = stat
        self.unit = unit


class MetricDataQueries:
    """
    # Arguments
    id : str
    metric_stat : MetricStat
    expression : str
    label : str
    return_data : boolean
    period : int
    """

    def __init__(
            self,
            id: str = none,
            metric_stat: MetricStat = none,
            expression: str = none,
            label: str = none,
            return_data: bool = none,
            period: int = none):
        self.id = id
        self.metric_stat = metric_stat
        self.expression = expression
        self.label = label
        self.return_data = return_data
        self.period = period


class CloudWatch:
    """
    # Arguments
    duration : str
    metric_data_queries : List[MetricDataQueries]
    """

    def __init__(
            self,
            duration: str = none,
            metric_data_queries: List[MetricDataQueries] = none):
        self.duration = duration
        self.metric_data_queries = metric_data_queries


class Headers:
    """
    # Arguments
    key : str
    value : str
    """

    def __init__(
            self,
            key: str = none,
            value: str = none):
        self.key = key
        self.value = value


class Web:
    """
    # Arguments
    method : str
    url : str
    headers : List[Headers]
    body : str
    timeout_seconds : int
    json_path : str
    insecure : bool
    """

    def __init__(
            self,
            method: str = none,
            url: str = none,
            headers: List[Headers] = none,
            body: str = none,
            timeout_seconds: int = none,
            json_path: str = none,
            insecure: bool = none):
        self.method = method
        self.url = url
        self.headers = headers
        self.body = body
        self.timeout_seconds = timeout_seconds
        self.json_path = json_path
        self.insecure = insecure


class Containers:
    """
    # Arguments
    name : str
    command : List[str]
    image : str
    """

    def __init__(
            self,
            name: str = none,
            command: List[str] = none,
            image: str = none):
        self.name = name
        self.command = command
        self.image = image


class TemplateSpec:
    """
    # Arguments
    containers : List[Containers]
    restart_policy : str
    """

    def __init__(
            self,
            containers: List[Containers] = none,
            restart_policy: str = none):
        self.containers = containers
        self.restart_policy = restart_policy


class JobSpec:
    """
    # Arguments
    backoff_limit : int
    template : TemplateSpec
    """

    def __init__(
            self,
            backoff_limit: int = none,
            template: TemplateSpec = none):
        self.backoff_limit = backoff_limit
        self.template = template


class Job:
    """
    # Arguments
    spec : JobSpec
    """

    def __init__(
            self,
            spec: JobSpec = none):
        self.spec = spec


class Provider:
    """
    # Arguments
    prometheus : Prometheus
    new_relic : NewRelic
    datadog : Datadog
    cloud_watch : CloudWatch
    web : Web
    job : Job
    """

    def __init__(
            self,
            prometheus: Prometheus = none,
            new_relic: NewRelic = none,
            datadog: Datadog = none,
            cloud_watch: CloudWatch = none,
            web: Web = none,
            job: Job = none):
        self.prometheus = prometheus
        self.new_relic = new_relic
        self.datadog = datadog
        self.cloud_watch = cloud_watch
        self.web = web
        self.job = job


class VerificationMetric:
    """
    # Arguments
    name : str
    dry_run : boolean
    interval : str
    initial_delay : str
    count : int
    success_condition : str
    failure_condition : str
    failure_limit : int
    consecutive_error_limit : int
    provider : Provider
    """

    def __init__(
            self,
            name: str = none,
            dry_run: bool = none,
            interval: str = none,
            initial_delay: str = none,
            count: int = none,
            success_condition: str = none,
            failure_condition: str = none,
            failure_limit: int = none,
            consecutive_error_limit: int = none,
            provider: Provider = none):
        self.name = name
        self.dry_run = dry_run
        self.interval = interval
        self.initial_delay = initial_delay
        self.count = count
        self.success_condition = success_condition
        self.failure_condition = failure_condition
        self.failure_limit = failure_limit
        self.consecutive_error_limit = consecutive_error_limit
        self.provider = provider
# endregion


# region Verification Template
class VerificationTemplate:
    """
    # Arguments
    name : str
    args : List[VerificationTemplateArgument]
    metrics : List[VerificationMetric]
    """

    def __init__(
            self,
            name: str = none,
            args: List[VerificationTemplateArgument] = none,
            metrics: List[VerificationMetric] = none,
            created_at: str = none,
            updated_at: str = none):
        self.name = name
        self.args = args
        self.metrics = metrics
# endregion


# region Client Requests
class CreateVerificationTemplateRequest:
    """
    # Arguments
    verification_template : VerificationTemplate
    """

    def __init__(self, verification_template: VerificationTemplate):
        self.verification_template = verification_template

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)
# endregion



import json
from enum import Enum
from typing import List

none = "d3043820717d74d9a17694c176d39733"


# region AutoScaler
class ResourceLimits:
    """
    # Arguments
    max_memory_gib: int
    max_v_cpu: int
    """

    def __init__(
            self,
            max_memory_gib: int = none,
            max_v_cpu: int = none):
        self.max_memory_gib = max_memory_gib
        self.max_v_cpu = max_v_cpu


class AgressiveScaleDown:
    """
    # Arguments
    is_enabled: bool
    """

    def __init__(
            self,
            is_enabled: bool = none):
        self.is_enabled = is_enabled


class Down:
    """
    # Arguments
    aggressive_scale_down: AgressiveScaleDown
    max_scale_down_percentage: int
    """

    def __init__(
            self,
            max_scale_down_percentage: int = none,
            aggressive_scale_down: AgressiveScaleDown = none):
        self.max_scale_down_percentage = max_scale_down_percentage
        self.aggressive_scale_down = aggressive_scale_down


class Headroom:
    """
    # Arguments
    cpu_per_unit: int
    memory_per_unit: int
    gpu_per_unit: int
    num_of_units: int
    """

    def __init__(
            self,
            cpu_per_unit: int = none,
            memory_per_unit: int = none,
            gpu_per_unit: int = none,
            num_of_units: int = none):
        self.cpu_per_unit = cpu_per_unit
        self.memory_per_unit = memory_per_unit
        self.gpu_per_unit = gpu_per_unit
        self.num_of_units = num_of_units


class AutoScaler:
    """
    # Arguments
    is_enabled: bool
    cooldown: int
    resource_limits: ResourceLimits
    down: Down
    headroom: Headroom
    is_auto_config: bool
    auto_headroom_percentage: int
    enable_automatic_and_manual_headroom: bool
    extended_resource_definitions: List[str]
    """

    def __init__(
            self,
            is_enabled: bool = none,
            cooldown: int = none,
            resource_limits: ResourceLimits = none,
            down: Down = none,
            headroom: Headroom = none,
            is_auto_config: bool = none,
            auto_headroom_percentage: int = none,
            enable_automatic_and_manual_headroom: bool = none,
            extended_resource_definitions: List[str] = none):
        self.is_enabled = is_enabled
        self.cooldown = cooldown
        self.resource_limits = resource_limits
        self.down = down
        self.headroom = headroom
        self.is_auto_config = is_auto_config
        self.auto_headroom_percentage = auto_headroom_percentage
        self.enable_automatic_and_manual_headroom = enable_automatic_and_manual_headroom
        self.extended_resource_definitions = extended_resource_definitions
# endregion


# region Capacity
class Capacity:
    """
    # Arguments
    minimum: int
    maximum: int
    target: int
    """

    def __init__(
            self,
            minimum: int = none,
            maximum: int = none,
            target: int = none):
        self.minimum = minimum
        self.maximum = maximum
        self.target = target
# endregion


# region Strategy
class AvailabilityVsCost(Enum):
    cost_oriented = "costOriented"
    balanced = "balanced"
    cheapest = "cheapest"


class ClusterOrientation:
    """
    # Arguments
    availability_vs_cost: AvailabilityVsCost
    """

    def __init__(
            self,
            availability_vs_cost: AvailabilityVsCost = none):
        self.availability_vs_cost = availability_vs_cost


class SpreadNodesBy(Enum):
    vcpu = "vcpu"
    count = "count"


class Strategy:
    """
    # Arguments
    utilize_reserved_instances: bool
    fallback_to_od: bool
    spot_percentage: int
    grace_period: int
    draining_timeout: int
    utilize_commitments: bool
    cluster_orientation: ClusterOrientation
    spread_nodes_by: SpreadNodesBy
    """

    def __init__(
            self,
            utilize_reserved_instances: bool = none,
            fallback_to_od: bool = none,
            spot_percentage: int = none,
            grace_period: int = none,
            draining_timeout: int = none,
            utilize_commitments: bool = none,
            cluster_orientation: ClusterOrientation = none,
            spread_nodes_by: SpreadNodesBy = none):
        self.utilize_reserved_instances = utilize_reserved_instances
        self.fallback_to_od = fallback_to_od
        self.spot_percentage = spot_percentage
        self.grace_period = grace_period
        self.draining_timeout = draining_timeout
        self.utilize_commitments = utilize_commitments
        self.cluster_orientation = cluster_orientation
        self.spread_nodes_by = spread_nodes_by
# endregion


# region Scheduling
class ClusterRoll:
    """
    # Arguments
    batch_min_healthy_percentage: int
    batch_size_percentage: int
    comment: str
    respect_pdb: bool
    """

    def __init__(
            self,
            batch_min_healthy_percentage: int = none,
            batch_size_percentage: int = none,
            comment: str = none,
            respect_pdb: bool = none):
        self.batch_min_healthy_percentage = batch_min_healthy_percentage
        self.batch_size_percentage = batch_size_percentage
        self.comment = comment
        self.respect_pdb = respect_pdb


class AmiAutoUpdate:
    """
    # Arguments
    patch: bool
    minor_version: bool
    apply_roll: bool
    cluster_roll: ClusterRoll
    """

    def __init__(
            self,
            patch: bool = none,
            minor_version: bool = none,
            apply_roll: bool = none,
            cluster_roll: ClusterRoll = none):
        self.patch = patch
        self.minor_version = minor_version
        self.apply_roll = apply_roll
        self.cluster_roll = cluster_roll


class Parameters:
    """
    # Arguments
    cluster_roll: ClusterRoll
    ami_auto_update: AmiAutoUpdate
    """

    def __init__(
            self,
            ami_auto_update: AmiAutoUpdate = none,
            cluster_roll: ClusterRoll = none):
        self.cluster_roll = cluster_roll
        self.ami_auto_update = ami_auto_update


class TaskType(Enum):
    cluster_roll = "clusterRoll"
    ami_auto_update = "amiAutoUpdate"


class Tasks:
    """
    # Arguments
    cron_expression: str
    is_enabled: bool
    parameters: Parameters
    task_type: TaskType
    """

    def __init__(
            self,
            cron_expression: str = none,
            is_enabled: bool = none,
            parameters: Parameters = none,
            task_type: TaskType = none):
        self.cron_expression = cron_expression
        self.is_enabled = is_enabled
        self.parameters = parameters
        self.task_type = task_type


class ShutdownHours:
    """
    # Arguments
    is_enabled: bool
    time_windows: List[str]
    """

    def __init__(
            self,
            is_enabled: bool = none,
            time_windows: List[str] = none):
        self.is_enabled = is_enabled
        self.time_windows = time_windows


class Scheduling:
    """
    # Arguments
    tasks: List[Tasks]
    shutdown_hours: ShutdownHours
    """

    def __init__(
            self,
            tasks: List[Tasks] = none,
            shutdown_hours: ShutdownHours = none):
        self.tasks = tasks
        self.shutdown_hours = shutdown_hours
# endregion


# region Security
class ContainerImage:
    """
    # Arguments
    approved_images: List[str]
    """

    def __init__(
            self,
            approved_images: List[str] = none):
        self.approved_images = approved_images


class Security:
    """
    # Arguments
    container_image: ContainerImage
    """

    def __init__(
            self,
            container_image: ContainerImage = none):
        self.container_image = container_image
# endregion


# region Compute
# region InstanceTypes
class Architectures(Enum):
    i386 = "i386"
    x86_64 = "x86_64"
    arm64 = "arm64"


class Categories(Enum):
    accelerated_computing = "Accelerated_computing"
    compute_optimized = "Compute_optimized"
    general_purpose = "General_purpose"
    memory_optimized = "Memory_optimized"
    storage_optimized = "Storage_optimized"


class DiskTypes(Enum):
    nvme = "NVMe"
    ebs = "EBS"
    ssd = "SSD"
    hdd = "HDD"


class Hypervisor(Enum):
    nitro = "nitro"
    xen = "xen"


class RootDeviceTypes(Enum):
    ebs = "ebs"
    instance_store = "instance-store"


class VirtualizationTypes(Enum):
    hvm = "hvm"
    paravirtual = "paravirtual"


class InstanceTypesFilters:
    """
    # Arguments
    architectures: List[Architectures]
    categories: List[Categories]
    disk_types: List[DiskTypes]
    exclude_families: List[str]
    exclude_metal: bool
    hypervisor: List[Hypervisor]
    include_families: List[str]
    is_ena_supported: bool
    max_gpu: int
    max_memory_gi_b: float
    max_network_performance: int
    max_vcpu: int
    min_enis: int
    min_gpu: int
    min_memory_gi_b: float
    min_network_performance: int
    min_vcpu: int
    root_device_types: List[RootDeviceTypes]
    virtualization_types: List[VirtualizationTypes]
    """

    def __init__(
            self,
            architectures: List[Architectures] = none,
            categories: List[Categories] = none,
            disk_types: List[DiskTypes] = none,
            exclude_families: List[str] = none,
            exclude_metal: bool = none,
            hypervisor: List[Hypervisor] = none,
            include_families: List[str] = none,
            is_ena_supported: bool = none,
            max_gpu: int = none,
            max_memory_gi_b: float = none,
            max_network_performance: int = none,
            max_vcpu: int = none,
            min_enis: int = none,
            min_gpu: int = none,
            min_memory_gi_b: float = none,
            min_network_performance: int = none,
            min_vcpu: int = none,
            root_device_types: List[RootDeviceTypes] = none,
            virtualization_types: List[VirtualizationTypes] = none):
        self.architectures = architectures
        self.categories = categories
        self.disk_types = disk_types
        self.exclude_families = exclude_families
        self.exclude_metal = exclude_metal
        self.hypervisor = hypervisor
        self.include_families = include_families
        self.is_ena_supported = is_ena_supported
        self.max_gpu = max_gpu
        self.max_memory_gi_b = max_memory_gi_b
        self.max_network_performance = max_network_performance
        self.max_vcpu = max_vcpu
        self.min_enis = min_enis
        self.min_gpu = min_gpu
        self.min_memory_gi_b = min_memory_gi_b
        self.min_network_performance = min_network_performance
        self.min_vcpu = min_vcpu
        self.root_device_types = root_device_types
        self.virtualization_types = virtualization_types


class InstanceTypes:
    """
    # Arguments
    blacklist: List[str]
    whitelist: List[str]
    filters: InstanceTypesFilters
    """

    def __init__(
            self,
            blacklist: List[str] = none,
            whitelist: List[str] = none,
            filters: InstanceTypesFilters = none):
        self.blacklist = blacklist
        self.whitelist = whitelist
        self.filters = filters
# endregion


# region LaunchSpecification
class DynamicVolumeSize:
    """
    # Arguments
    base_size: int
    resource: str
    size_per_resource_unit: int
    """

    def __init__(
            self,
            base_size: int = none,
            resource: str = none,
            size_per_resource_unit: int = none):
        self.base_size = base_size
        self.resource = resource
        self.size_per_resource_unit = size_per_resource_unit


class EBS:
    """
    # Arguments
    throughput: int
    delete_on_termination: bool
    encrypted: bool
    iops: int
    kms_key_id: str
    snapshot_id:  str
    volume_type: str
    volume_size: int
    dynamic_volume_size: DynamicVolumeSize
    """

    def __init__(
            self,
            throughput: int = none,
            delete_on_termination: bool = none,
            encrypted: bool = none,
            iops: int = none,
            kms_key_id: str = none,
            snapshot_id: str = none,
            volume_type: str = none,
            volume_size: int = none,
            dynamic_volume_size: DynamicVolumeSize = none):
        self.throughput = throughput
        self.delete_on_termination = delete_on_termination
        self.encrypted = encrypted
        self.iops = iops
        self.kms_key_id = kms_key_id
        self.snapshot_id = snapshot_id
        self.volume_type = volume_type
        self.volume_size = volume_size
        self.dynamic_volume_size = dynamic_volume_size


class BlockDeviceMappings:
    """
    # Arguments
    device_name: str
    ebs: EBS
    """

    def __init__(
            self,
            device_name: str = none,
            ebs: EBS = none):
        self.device_name = device_name
        self.ebs = ebs


class IamInstanceProfile:
    """
    # Arguments
    name: str
    arn: str
    """

    def __init__(
            self,
            name: str = none,
            arn: str = none):
        self.name = name
        self.arn = arn


class HttpEndpoint(Enum):
    disabled = "disabled"
    enabled = "enabled"


class HttpTokens(Enum):
    optional = "optional"
    required = "required"


class InstanceMetadataOptions:
    """
    # Arguments
    http_endpoint: HttpEndpoint
    http_put_response_hop_limit: int
    http_tokens: HttpTokens
    """

    def __init__(
            self,
            http_endpoint: HttpEndpoint = none,
            http_put_response_hop_limit: int = none,
            http_tokens: HttpTokens = none):
        self.http_endpoint = http_endpoint
        self.http_put_response_hop_limit = http_put_response_hop_limit
        self.http_tokens = http_tokens


class LoadBalancerType(Enum):
    classic = "CLASSIC"
    target_group = "TARGET_GROUP"


class LoadBalancer:
    """
    # Arguments
    arn: str
    name: str
    type: str
    """

    def __init__(
            self,
            arn: str = none,
            name: str = none,
            type: LoadBalancerType = none):
        self.arn = arn
        self.name = name
        self.type = type


class Tags:
    """
    # Arguments
    tag_key: str
    tag_value: str
    """

    def __init__(
            self,
            tag_key: str = none,
            tag_value: str = none):
        self.tag_key = tag_key
        self.tag_value = tag_value


class Volumes:
    """
    # Arguments
    should_tag: str
    """

    def __init__(
            self,
            should_tag: bool = none):
        self.should_tag = should_tag


class ResourceTagSpecification:
    """
    # Arguments
    volumes: Volumes
    """

    def __init__(
            self,
            volumes: Volumes = none):
        self.volumes = volumes


class LaunchSpecifications:
    """
    # Arguments
    associate_ipv6_address: bool
    associate_public_ip_address: bool
    block_device_mappings: List[BlockDeviceMappings]
    ebs_optimized: bool
    iam_instance_profile: IamInstanceProfile
    image_id: str
    instance_metadata_options: InstanceMetadataOptions
    key_pair: str
    load_balancers: List[LoadBalancers]
    monitoring: bool
    resource_tag_specification: ResourceTagSpecification
    root_volume_size: int
    security_group_ids: List[str]
    tags: List[Tags]
    use_as_template_only: bool
    user_data: str
    """

    def __init__(
            self,
            associate_ipv6_address: bool = none,
            associate_public_ip_address: bool = none,
            block_device_mappings: List[BlockDeviceMappings] = none,
            ebs_optimized: bool = none,
            iam_instance_profile: IamInstanceProfile = none,
            image_id: str = none,
            instance_metadata_options: InstanceMetadataOptions = none,
            key_pair: str = none,
            load_balancers: List[LoadBalancer] = none,
            monitoring: bool = none,
            resource_tag_specification: ResourceTagSpecification = none,
            root_volume_size: int = none,
            security_group_ids: List[str] = none,
            tags: List[Tags] = none,
            use_as_template_only: bool = none,
            user_data: str = none):
        self.associate_ipv6_address = associate_ipv6_address
        self.associate_public_ip_address = associate_public_ip_address
        self.block_device_mappings = block_device_mappings
        self.ebs_optimized = ebs_optimized
        self.iam_instance_profile = iam_instance_profile
        self.image_id = image_id
        self.instance_metadata_options = instance_metadata_options
        self.key_pair = key_pair
        self.load_balancers = load_balancers
        self.monitoring = monitoring
        self.resource_tag_specification = resource_tag_specification
        self.root_volume_size = root_volume_size
        self.security_group_ids = security_group_ids
        self.tags = tags
        self.use_as_template_only = use_as_template_only
        self.user_data = user_data
# endregion


class Compute:
    """
    # Arguments
    subnet_ids: List[str]
    instance_types: InstanceTypes
    launch_specification: LaunchSpecification
    """

    def __init__(
            self,
            subnet_ids: List[str] = none,
            instance_types: InstanceTypes = none,
            launch_specification: LaunchSpecifications = none):
        self.subnet_ids = subnet_ids
        self.instance_types = instance_types
        self.launch_specification = launch_specification
# endregion


# region Logging
class S3:
    """
    # Arguments
    id: str
    """

    def __init__(
            self,
            id: str = none):
        self.id = id


class Export:
    """
    # Arguments
    s3: S3
    """

    def __init__(
            self,
            s3: S3 = none):
        self.s3 = s3


class Logging:
    """
    # Arguments
    export: Export
    """

    def __init__(
            self,
            export: Export = none):
        self.export = export
# endregion


# region Ocean
class Ocean:
    """
    # Arguments
    name: str
    controller_cluster_id: str
    region: str
    auto_scaler: AutoScaler
    capacity: Capacity
    strategy: Strategy
    scheduling: Scheduling
    security: Security
    compute: Compute
    logging: Logging
    """

    def __init__(
            self,
            name: str = none,
            controller_cluster_id: str = none,
            region: str = none,
            auto_scaler: AutoScaler = none,
            capacity: Capacity = none,
            strategy: Strategy = none,
            scheduling: Scheduling = none,
            security: Security = none,
            compute: Compute = none,
            logging: Logging = none):
        self.name = name
        self.controller_cluster_id = controller_cluster_id
        self.region = region
        self.auto_scaler = auto_scaler
        self.capacity = capacity
        self.strategy = strategy
        self.scheduling = scheduling
        self.security = security
        self.compute = compute
        self.security = logging
# endregion


# region OceanRequest
class OceanRequest:
    def __init__(self, ocean: Ocean):
        self.cluster = ocean

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)
# endregion


class Type(Enum):
    label = "label"
    annotation = "annotation"


class Operator(Enum):
    equals = "equals"
    not_equals = "notEquals"
    exists = "exists"
    does_not_exist = "doesNotExist"


class Attribute:
    """
    # Arguments
    type: Type
    key: str
    operator: Operator
    value: str
    """

    def __init__(
            self,
            type: Type = none,
            key: str = none,
            operator: Operator = none,
            value: str = none):
        self.type = type
        self.key = key
        self.operator = operator
        self.value = value


class RightSizingRecommendationFilter:
    """
    # Attribute
    namespaces: List[str]
    attribute: Attribute
    """

    def __init__(
            self,
            namespaces: List[str] = none,
            attribute: Attribute = none):
        self.namespaces = namespaces
        self.attribute = attribute


class RightSizingRecommendationRequest:
    def __init__(self, filter: RightSizingRecommendationFilter = none):
        self.filter = filter

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class AllMatch:
    """
    # Arguments
    all_matches:  List[Attribute]
    """

    def __init__(
            self,
            all_matches:  List[Attribute] = none):
        self.all_matches = all_matches


class Conditions:
    """
    # Arguments
    any_match: List[AllMatch]
    """

    def __init__(
            self,
            any_match: List[AllMatch] = none):
        self.any_match = any_match


class Scope(Enum):
    namespace = "namespace"
    resource = "resource"


class Filter:
    """
    # Arguments
    conditions: Conditions
    scope: Scope
    """

    def __init__(
            self,
            conditions: Conditions = none):
        self.conditions = conditions


class GroupBy(Enum):
    namespace = "namespace"
    namespace_label = "namespace.label.${labelKey}"
    resource_label = "resource.label.${labelKey}"
    namespace_annotation = "namespace.annotation.${annotationKey}"
    resource_annotation = "resource.annotation.${annotationKey}"


class AggregatedClusterCosts:
    """
    # Arguments
    end_time: str
    aggregated_filter: Filter
    group_by: GroupBy
    start_time: str
    """

    def __init__(
            self,
            end_time: str = none,
            aggregated_filter: Filter = none,
            group_by: GroupBy = GroupBy.namespace.value,
            start_time: str = none):
        self.end_time = end_time
        self.aggregated_filter = aggregated_filter
        self.group_by = group_by
        self.start_time = start_time


class AggregatedClusterCostRequest:
    def __init__(self, aggregated_cluster_costs: AggregatedClusterCosts = none):
        self.end_time = aggregated_cluster_costs.end_time
        self.start_time = aggregated_cluster_costs.start_time
        self.group_by = aggregated_cluster_costs.group_by
        self.filter = aggregated_cluster_costs.aggregated_filter

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class Roll:
    """
    # Arguments
    batch_min_healthy_percentage: int
    batch_size_percentage: int
    comment: str
    disable_launch_spec_auto_scaling: bool
    instance_ids: List[str]
    launch_spec_ids: List[str]
    respect_pdb: bool
    """

    def __init__(
            self,
            batch_min_healthy_percentage: int = none,
            batch_size_percentage: int = none,
            comment: str = none,
            disable_launch_spec_auto_scaling: bool = none,
            launch_spec_ids: List[str] = none,
            instance_ids: List[str] = none,
            respect_pdb: bool = none):
        self.batch_min_healthy_percentage = batch_min_healthy_percentage
        self.batch_size_percentage = batch_size_percentage
        self.comment = comment
        self.disable_launch_spec_auto_scaling = disable_launch_spec_auto_scaling
        self.instance_ids = instance_ids
        self.launch_spec_ids = launch_spec_ids
        self.respect_pdb = respect_pdb

# region VirtualNodeGroup


class AutoScaleDown:
    """
    # Arguments
    max_scale_down_percentage: float
    """

    def __init__(
            self,
            max_scale_down_percentage: float = none):
        self.max_scale_down_percentage = max_scale_down_percentage


class AutoScale:
    """
    # Arguments
    auto_headroom_percentage: int
    down: class(AutoScaleDown)
    headrooms: List[Headroom]
    """

    def __init__(
            self,
            auto_headroom_percentage: int = none,
            down: AutoScaleDown = none,
            headrooms: List[Headroom] = none):
        self.auto_headroom_percentage = auto_headroom_percentage
        self.down = down
        self.headrooms = headrooms


class TagSelector:
    """
    # Arguments
    tag_key: str
    tag_value: str
    """

    def __init__(
            self,
            tag_key: str = none,
            tag_value: str = none):
        self.tag_key = tag_key
        self.tag_value = tag_value


class ElasticIpPool:
    """
    # Arguments
    tag_selector: TagSelector
    """

    def __init__(
            self,
            tag_selector: TagSelector = none):
        self.tag_selector = tag_selector


class ImageId:
    """
    # Arguments
    id: str = none
    """

    def __init__(
            self,
            id: str = none):
        self.id = id


class Labels:
    """
    # Arguments
    key: str
    value: str
    """

    def __init__(
            self,
            key: str = none,
            value: str = none):
        self.key = key
        self.value = value


class VNGResourceLimits:
    """
    # Arguments
    max_instance_count: int
    min_instance_count: int
    """

    def __init__(
            self,
            max_instance_count: int = none,
            min_instance_count: int = none):
        self.max_instance_count = max_instance_count
        self.min_instance_count = min_instance_count


class Config:
    """
    # Arguments
    headrooms: List[Headroom]
    """

    def __init__(
            self,
            headrooms: List[Headroom] = none):
        self.headrooms = headrooms


class VNGSchedulingTaskType(Enum):
    manual_headroom_update = "manualHeadroomUpdate"


class VNGSchedulingTasks:
    """
    # Arguments
    cron_expression: str
    is_enabled: bool
    parameters: Parameters
    task_type: TaskType
    """

    def __init__(
            self,
            cron_expression: str = none,
            is_enabled: bool = none,
            config: Config = none,
            task_type: VNGSchedulingTaskType = none):
        self.cron_expression = cron_expression
        self.is_enabled = is_enabled
        self.config = config
        self.task_type = task_type


class VNGSchedulingShutdownHours:
    """
    # Arguments
    is_enabled: bool
    time_windows: List[str]
    """

    def __init__(
            self,
            is_enabled: bool = none,
            time_windows: List[str] = none):
        self.is_enabled = is_enabled
        self.time_windows = time_windows


class VNGScheduling:
    """
    # Arguments
    tasks: List[VNGSchedulingTasks]
    shutdown_hours: VNGSchedulingShutdownHours
    """

    def __init__(
            self,
            tasks: List[VNGSchedulingTasks] = none,
            shutdown_hours: VNGSchedulingShutdownHours = none):
        self.tasks = tasks
        self.shutdown_hours = shutdown_hours


class VNGStrategy:
    """
    # Arguments
    spot_percentage: int
    utilizeReservedInstances : int
    fallbackToOd: bool
    spotInstancePools: int
    gracePeriod: int
    drainingTimeout: int
    utilizeCommitments: bool
    clusterOrientation: ClusterOrientation
    spreadNodesBy: SpreadNodesBy
    """

    def __init__(
            self,
            spot_percentage: int = none,
            utilize_reserved_instances: bool = none,
            fallback_to_od: bool = none,
            spot_instance_pools: int = none,
            grace_period: int = none,
            draining_timeout: int = none,
            utilize_commitments: bool = none,
            cluster_orientation: ClusterOrientation = none,
            spread_nodes_by: SpreadNodesBy = none):
        self.spot_percentage = spot_percentage
        self.utilize_reserved_instances = utilize_reserved_instances
        self.fallback_to_od = fallback_to_od
        self.spot_instance_pools = spot_instance_pools
        self.grace_period = grace_period
        self.draining_timeout = draining_timeout
        self.utilize_commitments = utilize_commitments
        self.cluster_orientation = cluster_orientation
        self.spread_nodes_by = spread_nodes_by


class Taints:
    """
    # Arguments
    tag_key: str
    tag_value: str
    """

    def __init__(
            self,
            effect: str = none,
            key: str = none,
            value: str = none):
        self.effect = effect
        self.key = key
        self.value = value


class VirtualNodeGroup:
    """
    # Arguments
    associate_public_ip_address: bool
    auto_scale: AutoScale
    block_device_mappings: List[BlockDeviceMappings]
    elastic_ip_pool: ElasticIpPool
    iam_instance_profile: IamInstanceProfile
    image_id: str
    images: List[ImageId]
    instance_metadata_options: InstanceMetadataOptions
    instance_types : List[str]
    instance_types_filters: InstanceTypesFilters
    labels: List[Labels]
    name: str
    ocean_id: str
    preferred_spot_types: List[str]
    resource_limits: VNGResourceLimits
    restrict_scale_down: bool
    root_volume_size: int
    scheduling: VNGScheduling
    strategy: VNGStrategy
    security_group_ids: List[str]
    subnet_ids: List[str]
    taints: List[Taints]
    user_data: str
    """

    def __init__(
            self,
            associate_public_ip_address: bool = none,
            auto_scale: AutoScale = none,
            block_device_mappings: List[BlockDeviceMappings] = none,
            elastic_ip_pool: ElasticIpPool = none,
            iam_instance_profile: IamInstanceProfile = none,
            image_id: str = none,
            images: List[ImageId] = none,
            instance_metadata_options: InstanceMetadataOptions = none,
            instance_types: List[str] = none,
            instance_types_filters: InstanceTypesFilters = none,
            labels: List[Labels] = none,
            name: str = none,
            ocean_id: str = none,
            preferred_spot_types: List[str] = none,
            resource_limits: VNGResourceLimits = none,
            restrict_scale_down: bool = none,
            root_volume_size: int = none,
            scheduling: VNGScheduling = none,
            strategy: VNGStrategy = none,
            security_group_ids: List[str] = none,
            subnet_ids: List[str] = none,
            tags: List[Tags] = none,
            taints: List[Taints] = none,
            user_data: str = none):
        self.associate_public_ip_address = associate_public_ip_address
        self.auto_scale = auto_scale
        self.block_device_mappings = block_device_mappings
        self.elastic_ip_pool = elastic_ip_pool
        self.iam_instance_profile = iam_instance_profile
        self.image_id = image_id
        self.images = images
        self.instance_metadata_options = instance_metadata_options
        self.instance_types = instance_types
        self.instance_types_filters = instance_types_filters
        self.labels = labels
        self.name = name
        self.ocean_id = ocean_id
        self.preferred_spot_types = preferred_spot_types
        self.resource_limits = resource_limits
        self.restrict_scale_down = restrict_scale_down
        self.root_volume_size = root_volume_size
        self.scheduling = scheduling
        self.strategy = strategy
        self.security_group_ids = security_group_ids
        self.subnet_ids = subnet_ids
        self.tags = tags
        self.taints = taints
        self.user_data = user_data
# endregion


# region Migration
class Migration:
    """
    # Arguments
    batch_size_percentage: int
    force_pod_eviction_on_pdb_failure: bool
    instance_ids: List[str]
    should_evict_stand_alone_pods: bool
    should_terminate_drained_nodes: bool
    state: str
    """

    def __init__(
            self,
            batch_size_percentage: int = none,
            force_pod_eviction_on_pdb_failure: bool = none,
            instance_ids: List[str] = none,
            should_evict_stand_alone_pods: bool = none,
            should_terminate_drained_nodes: bool = none,
            state: str = none):
        self.batch_size_percentage = batch_size_percentage
        self.force_pod_eviction_on_pdb_failure = force_pod_eviction_on_pdb_failure
        self.instance_ids = instance_ids
        self.should_evict_stand_alone_pods = should_evict_stand_alone_pods
        self.should_terminate_drained_nodes = should_terminate_drained_nodes
        self.state = state
# endregion


# region DetachInstances
class DetachInstances:
    """
    # Arguments
    instances_to_detach: List[str]
    should_decrement_target_capacity: bool
    should_terminate_instances: bool
    """

    def __init__(
            self,
            instances_to_detach: List[str] = none,
            should_decrement_target_capacity: bool = none,
            should_terminate_instances: bool = none):
        self.instances_to_detach = instances_to_detach
        self.should_decrement_target_capacity = should_decrement_target_capacity
        self.should_terminate_instances = should_terminate_instances
# endregion

# region ImportASGToOcean


class ImportASGToOceanCluster:
    """
    # Arguments
    instance_types: List[str]
    """

    def __init__(
            self,
            instance_types: List[str] = none):
        self.instance_types = instance_types
# endregion


# region ImportASGToOceanVNG
class ImportASGToOceanVNG:
    """
    # Arguments
    labels: List[Labels]
    name: str
    taints: List[Taints]
    """

    def __init__(
            self,
            name: str = none,
            labels: List[Labels] = none,
            taints: List[Taints] = none):
        self.name = name
        self.labels = labels
        self.taints = taints
# endregion


# region ImportEKSNodeGroupToOceanVNG
class ImportEKSNodeGroupToOceanVNG:
    """
    # Arguments
    labels: List[Labels]
    name: str
    """

    def __init__(
            self,
            name: str = none,
            labels: List[Labels] = none):
        self.name = name
        self.labels = labels
# endregion


# region ExtendedResourceDefinition
class ExtendedResourceDefinition:
    """
    # Arguments
    mapping: dict
    name: str
    """

    def __init__(
            self,
            mapping: dict = none,
            name: str = none):
        self.mapping = mapping
        self.name = name
# endregion


class VNGRequest:
    def __init__(self, vng: VirtualNodeGroup):
        self.launch_spec = vng

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class ImportASGToOceanClusterRequest:
    def __init__(self, cluster: ImportASGToOceanCluster):
        self.cluster = cluster

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class ImportASGToOceanVNGRequest:
    def __init__(self, launch_spec: ImportASGToOceanVNG):
        self.launch_spec = launch_spec

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class ImportEKSNodeGroupToOceanVNGRequest:
    def __init__(self, launch_spec: ImportEKSNodeGroupToOceanVNG):
        self.launch_spec = launch_spec

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class DetachInstancesRequest:
    def __init__(self, detachInstances: DetachInstances):
        self.instances_to_detach = detachInstances.instances_to_detach
        self.should_decrement_target_capacity = detachInstances.should_decrement_target_capacity
        self.should_terminate_instances = detachInstances.should_terminate_instances

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class LoadBalancersRequest:
    def __init__(self, load_balancers: List[LoadBalancer]):
        self.load_balancers = load_balancers

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class ClusterRollInitiateRequest:
    def __init__(self, roll: Roll = none):
        self.roll = roll

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class ClusterRollUpdateRequest:
    def __init__(self, status: str = none):
        self.roll = dict(status=status)

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class InstanceTypesFilterSimulationRequest:
    def __init__(self, instance_type_filter: InstanceTypesFilters = none):
        self.filters = instance_type_filter

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class LaunchNodesRequest:
    def __init__(self, amount: int = none):
        self.launch_request = dict(amount=amount)

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class MigrationRequest:
    def __init__(self, migration: Migration):
        self.migration = migration

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class ExtendedResourceDefinitionRequest:
    def __init__(self, extended_resource_defintion: ExtendedResourceDefinition):
        self.extended_resource_defintion = extended_resource_defintion

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class InstanceTypesFilterSimulationRequestForVNG:
    def __init__(self, instance_types_filters: InstanceTypesFilters = none):
        self.instance_types_filters = instance_types_filters

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)



import json
from enum import Enum
from typing import List

none = "d3043820717d74d9a17694c176d39733"


# region WeeklyRepetitionBasis
class IntervalHours:
    """
    # Arguments
    start_time: str
    end_time: str
    """

    def __init__(
            self,
            start_time: str = none,
            end_time: str = none
    ):
        self.start_time = start_time
        self.end_time = end_time


class IntervalDays(Enum):
    Monday = "MONDAY"
    Tuesday = "TUESDAY"
    Wednesday = "WEDNESDAY"
    Thursday = "THURSDAY"
    Friday = "FRIDAY"
    Saturday = "SATURDAY"
    Sunday = "SUNDAY"


class WeeklyRepetitionBasis:
    """
    # Arguments
    interval_days: List[IntervalDays]
    interval_hours: IntervalHours
    """

    def __init__(
            self,
            interval_days: List[IntervalDays] = none,
            interval_hours: IntervalHours = none
    ):
        self.interval_days = interval_days
        self.interval_hours = interval_hours
# endregion

# region MonthlyRepetitionBasis


class WeekOfTheMonth(Enum):
    First = "FIRST"
    Second = "SECOND"
    Third = "THIRD"
    Fourth = "FOURTH"
    Last = "LAST"


class MonthlyRepetitionBasis:
    """
    # Arguments
    interval_months: List[int]
    week_of_the_month: List[WeekOfTheMonth]
    weekly_repetition_basis: WeeklyRepetitionBasis
    """

    def __init__(
            self,
            interval_months: List[int] = none,
            week_of_the_month: List[WeekOfTheMonth] = none,
            weekly_repetition_basis: WeeklyRepetitionBasis = none
    ):
        self.interval_months = interval_months
        self.week_of_the_month = week_of_the_month
        self.weekly_repetition_basis = weekly_repetition_basis
# endregion


# region RecommendationApplicationInterval
class RepetitionBasis(Enum):
    Weekly = "WEEKLY"
    Monthly = "MONTHLY"


class RecommendationApplicationInterval:
    """
    # Arguments
    repetition_basis: RepetitionBasis
    monthly_repetition_basis: MonthlyRepetitionBasis
    weekly_repetition_basis: WeeklyRepetitionBasis
    """

    def __init__(
            self,
            repetition_basis: RepetitionBasis = none,
            monthly_repetition_basis: MonthlyRepetitionBasis = none,
            weekly_repetition_basis: WeeklyRepetitionBasis = none
    ):
        self.repetition_basis = repetition_basis
        self.monthly_repetition_basis = monthly_repetition_basis
        self.weekly_repetition_basis = weekly_repetition_basis
# endregion

# region RecommendationApplicationMinThreshold


class RecommendationApplicationMinThreshold:
    """
    # Arguments
    cpu_percentage: float
    memory_percentage: float
    """

    def __init__(
            self,
            cpu_percentage: float = none,
            memory_percentage: float = none
    ):
        self.cpu_percentage = cpu_percentage
        self.memory_percentage = memory_percentage
# endregion

# region RecommendationApplicationBoundaries


class CPU:
    """
    # Arguments
    min: int
    max: int
    """

    def __init__(
            self,
            min: int = none,
            max: int = none
    ):
        self.min = min
        self.max = max


class Memory:
    """
    # Arguments
    min: int
    max: int
    """

    def __init__(
            self,
            min: int = none,
            max: int = none
    ):
        self.min = min
        self.max = max


class RecommendationApplicationBoundaries:
    """
    # Arguments
    cpu: CPU
    memory: Memory
    """

    def __init__(
            self,
            cpu: CPU = none,
            memory: Memory = none
    ):
        self.cpu = cpu
        self.memory = memory
# endregion

# region RecommendationApplicationOverheadValues


class RecommendationApplicationOverheadValues:
    """
    # Arguments
    cpu_percentage: float
    memory_percentage: float
    """

    def __init__(
            self,
            cpu_percentage: float = none,
            memory_percentage: float = none
    ):
        self.cpu_percentage = cpu_percentage
        self.memory_percentage = memory_percentage
# endregion


# region RecommendationApplicationHPA
class RecommendationApplicationHPA:
    """
    # Arguments
    allow_h_p_a_recommendations: bool
    """

    def __init__(
            self,
            allow_h_p_a_recommendations: bool = none,
    ):
        self.allow_h_p_a_recommendations = allow_h_p_a_recommendations

# endregion


# region Label

class Label:
    """
    # Arguments
    key: str
    value: str
    """

    def __init__(
            self,
            key: str = none,
            value: str = none
    ):
        self.key = key
        self.value = value
# endregion

# region Workload


class Workload:
    """
    # Arguments
    name: str
    workload_type: str
    regex_name: str
    """

    def __init__(
            self,
            name: str = none,
            workload_type: str = none,
            regex_name: str = none

    ):
        self.name = name
        self.workload_type = workload_type
        self.regex_name = regex_name

# endregion


# region Namespace


class Namespace:
    """
    # Arguments
    namespace_name: str
    labels: List[Label]
    workloads: List[Workload]
    """

    def __init__(
            self,
            namespace_name: str = none,
            labels: List[Label] = none,
            workloads: List[Workload] = none
    ):
        self.namespace_name = namespace_name
        self.labels = labels
        self.workloads = workloads
# endregion


class ClusterResources:
    """
    # Arguments
    data: dict
    """

    def __init__(
            self,
            data: dict
    ):
        self.data = data

# endregion


class ClusterLabels:
    """
    # Arguments
    data: dict
    """

    def __init__(
            self,
            data: dict
    ):
        self.data = data

# endregion


class RestartReplicas(Enum):
    More_than_one_replica = "MORE_THAN_ONE_REPLICA"
    All_manifest = "ALL_MANIFEST"
    No_restart = "NO_RESTART"


class CreateRightSizingRuleRequest:
    def __init__(self,
                 rule_name: str,
                 restart_replicas: RestartReplicas,
                 exclude_preliminary_recommendations: bool,
                 recommendation_application_intervals: List[RecommendationApplicationInterval],
                 recommendation_application_min_threshold: RecommendationApplicationMinThreshold,
                 recommendation_application_boundaries: RecommendationApplicationBoundaries,
                 recommendation_application_overhead_values: RecommendationApplicationOverheadValues,
                 recommendation_application_hpa: RecommendationApplicationHPA):
        self.rule_name = rule_name
        self.restart_replicas = restart_replicas
        self.exclude_preliminary_recommendations = exclude_preliminary_recommendations
        self.recommendation_application_intervals = recommendation_application_intervals
        self.recommendation_application_min_threshold = recommendation_application_min_threshold
        self.recommendation_application_boundaries = recommendation_application_boundaries
        self.recommendation_application_overhead_values = recommendation_application_overhead_values
        self.recommendation_application_h_p_a = recommendation_application_hpa

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class DeleteRightSizingRulesRequest:
    def __init__(self, rule_names: List[str]):
        self.rule_names = rule_names

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class UpdateRightSizingRuleRequest:
    def __init__(self,
                 restart_replicas: RestartReplicas,
                 exclude_preliminary_recommendations: bool,
                 recommendation_application_intervals: List[RecommendationApplicationInterval],
                 recommendation_application_min_threshold: RecommendationApplicationMinThreshold,
                 recommendation_application_boundaries: RecommendationApplicationBoundaries,
                 recommendation_application_overhead_values: RecommendationApplicationOverheadValues,
                 recommendation_application_hpa: RecommendationApplicationHPA):
        self.restart_replicas = restart_replicas
        self.exclude_preliminary_recommendations = exclude_preliminary_recommendations
        self.recommendation_application_intervals = recommendation_application_intervals
        self.recommendation_application_min_threshold = recommendation_application_min_threshold
        self.recommendation_application_boundaries = recommendation_application_boundaries
        self.recommendation_application_overhead_values = recommendation_application_overhead_values
        self.recommendation_application_h_p_a = recommendation_application_hpa

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class AttachRightSizingRuleRequest:
    def __init__(self, namespaces: List[Namespace]):
        self.namespaces = namespaces

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class DetachRightSizingRuleRequest:
    def __init__(self, namespaces: List[Namespace]):
        self.namespaces = namespaces

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class GetOceanRightSizingRecommendationsRequest:
    def __init__(self, cluster_resources: ClusterResources, cluster_labels: ClusterLabels):
        if cluster_resources is not None:
            self.cluster_resources = cluster_resources.data
        if cluster_labels is not None:
            self.cluster_labels = cluster_labels.data

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)



import json
from typing import List

none = "d3043820717d74d9a17694c176d39733"

# region organization


class Statement:
    """
    # Arguments
    effect: str
    actions: List[str]
    resources: List[str]
    """

    def __init__(
            self,
            effect: str = none,
            actions: List[str] = None,
            resources: List[str] = None):

        self.effect = effect
        self.actions = actions
        self.resources = resources


class PolicyContent:
    """
    # Arguments
    statements: List[Statement]
    """

    def __init__(
            self,
            statements: List[Statement] = none):

        self.statements = statements


class AccessPolicy:
    """
        # Arguments
        description: str
        name: str
        type: str
        policy_content: PolicyContent
        """

    def __init__(
            self,
            description: str = none,
            name: str = none,
            type: str = none,
            policy_content: PolicyContent = none):

        self.description = description
        self.name = name
        self.type = type
        self.policy_content = policy_content


class PolicyMapping:
    """
    # Arguments
    account_ids: List[str]
    policy_id: str
    """

    def __init__(
            self,
            policy_id: str = none,
            account_ids: List[str] = none):

        self.policy_id = policy_id
        self.account_ids = account_ids


class UserGroup:
    """
        # Arguments
        description: str
        name: str
        policies: List[PolicyMapping]
        user_ids: List[str]
        """

    def __init__(
            self,
            description: str = none,
            name: str = none,
            policies: List[PolicyMapping] = none,
            user_ids: List[str] = none):

        self.description = description
        self.name = name
        self.policies = policies
        self.user_ids = user_ids

# endregion


class AccessPolicyCreationRequest:
    def __init__(self, policy: AccessPolicy):
        self.policy = policy

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class AccessPolicyUpdationRequest:
    def __init__(self, policy: AccessPolicy):
        self.policy = policy

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


class UserGroupCreationRequest:
    def __init__(self, group: UserGroup):
        self.description = group.description
        self.name = group.name
        self.policies = group.policies
        self.user_ids = group.user_ids

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)



import json
import os
import re

import logging
import requests

import urllib3.util

from .version import __version__

VAR_SPOTINST_LOG_LEVEL = 'SPOTINST_LOG_LEVEL'


version = {'__version__': __version__}

_SpotinstClient__spotinst_sdk_python_agent_name = 'spotinst-sdk-python'
_SpotinstClient__spotinst_sdk_user_agent = '{}/{}'.format(
    _SpotinstClient__spotinst_sdk_python_agent_name, version['__version__'])


class Client:
    camel_pat = re.compile(r'([A-Z])')
    under_pat = re.compile(r'_([a-z])')

    __account_id_key = "accountId"

    def __init__(self, session=None,
                 print_output=True,
                 log_level=None,
                 user_agent=None,
                 timeout=None):

        self.base_url = session.base_url
        self.auth_token = session.auth_token
        self.account_id = session.account_id

        self.should_print_output = print_output
        self.user_agent = user_agent

        # initialize logger
        self.logger = self.init_logger()
        self.set_log_level(log_level=log_level)

        self.timeout = timeout

    def validate_url(self, url):
        self.print_output("Input Url - " + self.base_url + url)
        self.print_output("Parsed Url - " +
                          urllib3.util.parse_url(self.base_url + url).url)
        if urllib3.util.parse_url(self.base_url + url).url == self.base_url + url:
            return self.base_url + url
        else:
            raise Exception("UNSAFE_URL")

    def send_get(self, url, entity_name, query_params=None):
        agent = self.resolve_user_agent()

        if query_params is not None:
            query_params = self.build_query_params_with_input(query_params)
        else:
            query_params = self.build_query_params()

        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )

        self.print_output("Sending get request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))

        result = requests.get(
            self.validate_url(url), params=query_params, headers=headers, timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            data = json.loads(result.content.decode('utf-8'))
            return data
        else:
            self.handle_exception("getting {}".format(entity_name), result)

    def send_delete(self, url, entity_name):
        agent = self.resolve_user_agent()
        query_params = self.build_query_params()
        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )

        self.print_output("Sending deletion request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))

        result = requests.delete(
            self.validate_url(url), params=query_params, headers=headers, timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            return True
        else:
            self.handle_exception("deleting {}".format(entity_name), result)

    def send_delete_with_body(self, body, url, entity_name):
        agent = self.resolve_user_agent()
        query_params = self.build_query_params()
        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )

        self.print_output("Sending deletion request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))
        self.print_output("Request Body - " + str(body))

        result = requests.delete(
            self.validate_url(url),
            params=query_params,
            headers=headers,
            data=body,
            timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            return True
        else:
            self.handle_exception("deleting {}".format(entity_name), result)

    def send_delete_with_params(self, url, entity_name, user_query_params):
        agent = self.resolve_user_agent()

        query_params = self.build_query_params_with_input(user_query_params)

        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )
        self.print_output("Sending deletion request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))

        result = requests.delete(
            self.validate_url(url),
            params=query_params,
            headers=headers,
            timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            return True
        else:
            self.handle_exception("deleting {}".format(entity_name), result)

    def send_post(self, url, entity_name, body=None, query_params=None):
        agent = self.resolve_user_agent()

        if query_params is not None:
            query_params = self.build_query_params_with_input(query_params)
        else:
            query_params = self.build_query_params()

        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )

        self.print_output("Sending post request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))
        self.print_output("Request Body - " + str(body))

        result = requests.post(
            self.validate_url(url),
            params=query_params,
            data=body,
            headers=headers,
            timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            data = json.loads(result.content.decode('utf-8'))
            return data
        else:
            self.handle_exception("creating {}".format(entity_name), result)

    def send_post_with_params(self, url, entity_name, body, user_query_params):
        agent = self.resolve_user_agent()

        query_params = self.build_query_params_with_input(user_query_params)

        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )

        self.print_output("Sending post request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))
        self.print_output("Request Body - " + str(body))

        result = requests.post(
            self.validate_url(url),
            params=query_params,
            data=body,
            headers=headers,
            timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            data = json.loads(result.content.decode('utf-8'))
            return data
        else:
            self.handle_exception("creating {}".format(entity_name), result)

    def send_put(self, url, entity_name, query_params=None, body=None):
        agent = self.resolve_user_agent()

        if query_params is not None:
            query_params = self.build_query_params_with_input(query_params)
        else:
            query_params = self.build_query_params()

        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )

        self.print_output("Sending put request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))
        self.print_output("Request Body - " + str(body))

        result = requests.put(
            self.validate_url(url),
            params=query_params,
            data=body,
            headers=headers,
            timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            data = json.loads(result.content.decode('utf-8'))
            return data
        else:
            self.handle_exception("updating {}".format(entity_name), result)

    def send_put_with_params(self, body, url, entity_name, user_query_params):
        agent = self.resolve_user_agent()
        query_params = self.build_query_params_with_input(user_query_params)

        headers = dict(
            {
                'User-Agent': agent,
                'Content-Type': 'application/json',
                'Authorization': 'Bearer ' + self.auth_token
            }
        )

        self.print_output("Sending put request to spotinst API.")
        self.print_output("Request Query Params - " + str(query_params))
        self.print_output("Request Body - " + str(body))

        result = requests.put(
            self.validate_url(url),
            params=query_params,
            data=body,
            headers=headers,
            timeout=self.timeout)

        if result.status_code == requests.codes['ok']:
            self.print_output("Success")
            self.print_output("Response - " + str(result.json()))
            data = json.loads(result.content.decode('utf-8'))
            return data
        else:
            self.handle_exception("updating {}".format(entity_name), result)

    def resolve_user_agent(self):
        global _SpotinstClient__spotinst_sdk_user_agent
        agent = _SpotinstClient__spotinst_sdk_user_agent
        if self.user_agent is not None:
            agent = '{}+{}'.format(self.user_agent, agent)
        return agent

    def handle_exception(self, action_string, result):
        self.print_output(result.status_code)

        if result.content == "Bad Request":
            data = dict(response=result.content)
        else:
            data = json.loads(result.content.decode('utf-8'))

        response_json = json.dumps(data["response"])
        self.print_output(response_json)

        raise SpotinstClientException(
            "Error encountered while " +
            action_string,
            response_json)

    def convert_json(self, val, convert):
        new_json = {}
        if val is None:
            return val
        elif type(val) in (int, float, bool, "".__class__, u"".__class__):
            return val

        for k, v in list(val.items()):
            new_v = v
            if isinstance(v, dict):
                new_v = self.convert_json(v, convert)
            elif isinstance(v, list):
                new_v = list()
                for x in v:
                    new_v.append(self.convert_json(x, convert))
            new_json[convert(k)] = new_v
        return new_json

    def convert_json_with_list_of_lists(self, val, convert):
        new_json = {}
        if val is None:
            return val
        elif type(val) in (int, float, bool, "".__class__, u"".__class__):
            return val

        if isinstance(val, list):
            new_v = list()
            for x in val:
                new_v.append(self.convert_json_with_list_of_lists(x, convert))
            return new_v
        else:
            for k, v in list(val.items()):
                new_v = v
                if isinstance(v, dict):
                    new_v = self.convert_json_with_list_of_lists(v, convert)
                elif isinstance(v, list):
                    new_v = list()
                    for x in v:
                        new_v.append(
                            self.convert_json_with_list_of_lists(x, convert))
                new_json[convert(k)] = new_v
        return new_json

    def exclude_missing(self, obj):
        # Delete keys with the value 'none' in a dictionary, recursively.

        # if obj.items() is not None:
        if obj.items() is not None:
            for key, value in list(obj.items()):

                # Remove none values
                if value == "d3043820717d74d9a17694c176d39733":
                    del obj[key]

                # Handle Objects
                elif isinstance(value, dict):
                    self.exclude_missing(obj=value)

                # Handle lists
                elif self.is_sequence(arg=value):
                    for listitem in value:
                        # Handle Lists of objects
                        try:
                            self.exclude_missing(obj=listitem)
                        except AttributeError:
                            pass
        return obj  # For convenience

    def is_sequence(self, arg):
        return (not hasattr(arg, "strip") and
                hasattr(arg, "__getitem__") or
                hasattr(arg, "__iter__"))

    def build_query_params(self):
        query_params = None
        if self.account_id is not None:
            query_params = dict({self.__account_id_key: self.account_id})

        return query_params

    def build_query_params_with_input(self, user_params):
        query_params = dict()
        if self.account_id is not None:
            query_params = dict({self.__account_id_key: self.account_id})

        if user_params is not None:
            query_params = self.merge_two_dicts(query_params, user_params)

        return query_params

    def print_output(self, output, level="debug"):
        if self.should_print_output is True:
            if level == "debug":
                self.logger.debug(output)
            if level == "info":
                self.logger.info(output)
            if level == "warn":
                self.logger.warning(output)
            if level == "error":
                self.logger.error(output)
            if level == "critical":
                self.logger.critical(output)

    @staticmethod
    def init_logger():
        logging.basicConfig(level=logging.CRITICAL)
        logger = logging.getLogger(__name__)
        if not logger.hasHandlers():
            handler = logging.StreamHandler()
            formatter = logging.Formatter(
                '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        return logger

    def set_log_level(self, log_level):
        if log_level is None:
            level = os.environ.get(VAR_SPOTINST_LOG_LEVEL, 'critical')
        else:
            level = log_level

        if level == "debug":
            self.logger.setLevel(logging.DEBUG)
        if level == "info":
            self.logger.setLevel(logging.INFO)
        if level == "warn":
            self.logger.setLevel(logging.WARN)
        if level == "error":
            self.logger.setLevel(logging.ERROR)
        if level == "critical":
            self.logger.setLevel(logging.CRITICAL)

    @staticmethod
    def merge_two_dicts(x, y):
        z = x.copy()  # start with x's keys and values
        z.update(y)  # modifies z with y's keys and values & returns None
        return z

    def camel_to_underscore(self, name):
        return self.camel_pat.sub(lambda x: '_' + x.group(1).lower(), name)

    def underscore_to_camel(self, name):
        return self.under_pat.sub(lambda x: x.group(1).upper(), name)


class SpotinstClientException(Exception):
    def __init__(self, message, response):
        self.message = message + "\n" + response
        # Call the base class constructor with the parameters it needs
        super(SpotinstClientException, self).__init__(message)
    # endregion



import json
from typing import List

from spotinst_sdk2.client import Client
import spotinst_sdk2.models.admin.user_mapping as spotinst_user_mapping
import spotinst_sdk2.models.admin.organization as admin_org


class AdminClient(Client):
    __base_setup_url = "/setup"

    # region Organization and Account
    def create_organization(self, org_name: str):
        """
        Create an organization 

        # Arguments
        org_name (String): Orgnanization name

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_post(
            url=self.__base_setup_url +
            "/organization",
            body=json.dumps(dict(organization=dict(name=org_name))),
            entity_name="organization"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"][0]

    def delete_organization(self, org_id: str):
        """
        delete organization 

        # Arguments
        org_id (String): Organization Id

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_delete(
            url=self.__base_setup_url +
            "/organization/" + str(org_id),
            entity_name="organization"
        )

        return response

    def create_aws_external_id(self):
        """
        Important note: This is deprecated, please use setup_aws client instead(SetupAWSClient#create_external_id)

        Create aws account external id.
        You should use the external id when creating your AWS role for your spot account

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_post(
            url=self.__base_setup_url + "/credentials/aws/externalId",
            entity_name="credentials"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"][0]

    def set_cloud_credentials(self, iam_role: str, external_id: str = None):
        """
        Important note: This is deprecated, please use setup_aws client instead(SetupAWSClient#set_credentials)

        set cloud credentials 
        Please create external id using spot api (see #AdminClient.create_aws_external_id)
        and use it when creating the AWS role

        # Arguments
        iam_role (String): IAM Role
        external_id (String) (Optional): External ID

        # Returns
        (Object): Spotinst API response 
        """
        credentials = {"iamRole": iam_role}

        if external_id is not None:
            credentials['externalId'] = external_id

        response = self.send_post(
            url=self.__base_setup_url +
            "/credentials/aws",
            body=json.dumps(dict(credentials=credentials)),
            entity_name="credentials"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def create_account(self, account_name: str):
        """
        create an account 

        # Arguments
        account_name (String): Account Name

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_post(
            url=self.__base_setup_url +
            "/account",
            body=json.dumps(dict(account=dict(name=account_name))),
            entity_name="account"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"][0]

    def update_account(self, account_id: str, new_account_name: str, slack_notification_channels: List[str] = None):
        """
        create an account 

        # Arguments
        account_id (String): Account Id
        new_account_name (String): New Account Name
        slack_notification_channels List(str): List of slack notification channels

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_put(
            url=self.__base_setup_url +
            "/account/" + account_id,
            body=json.dumps(dict(account=dict(
                name=new_account_name, slackNotificationChannels=slack_notification_channels))),
            entity_name="account"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def get_accounts(self):
        """
        get accounts in organization

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_get(
            url=self.__base_setup_url +
            "/account",
            entity_name="account"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"]

    def delete_account(self, account_name: str):
        """
        delete account

        # Arguments
        account_name (String): Account Name

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_delete(
            url=self.__base_setup_url +
            "/account/" + account_name,
            entity_name="account"
        )

        return response

    def create_user(self, first_name: str, last_name: str, email: str, password: str, role: str):
        """
        Create user

        # Arguments
        first_name (String): Users first name
        last_name (String): User last name
        email (String): Eser email
        password (String): User email
        role (String): User role

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_post(
            url=self.__base_setup_url +
            "/user",
            body=json.dumps(dict(
                firstName=first_name,
                lastName=last_name,
                email=email,
                password=password,
                role=role)),
            entity_name="user"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"][0]

    def add_existing_user(self, user_email, role):
        """
        Add existing user

        # Arguments
        user_email (String): User email
        role (String): User role

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_post(
            url=self.__base_setup_url +
            "/account/" + self.account_id +
            "/user",
            body=json.dumps(dict(userEmail=user_email, role=role)),
            entity_name="user"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def update_user_role(self, user_email, role):
        """
        Update existing user

        # Arguments
        user_email (String): User email
        role (String): User role

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_put(
            url=self.__base_setup_url +
            "/account/" + self.account_id +
            "/user",
            body=json.dumps(dict(userEmail=user_email, role=role)),
            entity_name="user"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def detach_user(self, user_email: str):
        """
        Delete existing user

        # Arguments
        user_email (String): User email

        # Returns
        (Object): Spotinst API response 
        """
        response = self.send_delete_with_body(
            url=self.__base_setup_url +
            "/account/" + self.account_id +
            "/user",
            body=json.dumps(dict(userEmail=user_email)),
            entity_name="user"
        )

        return response

    def get_user(self, user_email: str):
        """
        Get user

        # Arguments
        user_email (String): User email

        # Returns
        (Object): Spotinst API response 
        """
        query_params = dict(userEmail=user_email)
        response = self.send_get(
            url=self.__base_setup_url + "/accountUserMapping",
            query_params=query_params,
            entity_name="user"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"]

    def assign_user_to_account(self, mappings):
        """
        Assign user to account

        # Arguments
        mappings (List): List of UserMapping Objects

        # Returns
        (Object): Spotinst API response 
        """
        mappings = spotinst_user_mapping.UserMappingRequest(mappings)

        excluded_group_dict = self.exclude_missing(
            json.loads(mappings.toJSON()))

        formatted_group_dict = self.convert_json(
            excluded_group_dict, self.underscore_to_camel)

        body_json = json.dumps(formatted_group_dict)

        response = self.send_post(
            url=self.__base_setup_url + "/accountUserMapping",
            body=body_json,
            entity_name="user"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    # endregion

    def get_users(self):
        """
        Retrieves all users from an organization.

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_get(
            url=self.__base_setup_url + "/organization/user", entity_name="user"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"]

    def get_policies(self):
        """
        Retrieves all policies from an organization.

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_get(
            url=self.__base_setup_url + "/organization/policy", entity_name="policy"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"]

    def get_user_details(self, user_id: str):
        """
        Retrieves an individual user details.

        # Arguments
        user_id (String): User ID

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_get(
            url=self.__base_setup_url + "/user/" + user_id, entity_name="user"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"][0]

    def delete_user(self, user_id: str):
        """
        Deletes a user (console or programmatic) from an organization.

        # Arguments
        user_id (String): User ID

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_delete(
            url=self.__base_setup_url + "/user/" + user_id, entity_name="user"
        )

        return response

    def update_user_to_user_group_mapping(self, user_id: str, user_group_ids: List[str]):
        """
        Update the mapping of a given user to user groups

        # Arguments
        user_id (String): Identifier of a user.
        user_group_ids (List): A list of the user groups to register the given user to

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_put(
            url=self.__base_setup_url + "/user/" + user_id + "/userGroupMapping",
            body=json.dumps(dict(userGroupIds=user_group_ids)),
            entity_name="user",
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def update_user_to_policy_mapping(self, user_id: str, policies):
        """
        Update the mapping of a given user to policies

        # Arguments
        user_id (String): Identifier of a user.
        policies (List): A list of policies to register under the given user

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_put(
            url=self.__base_setup_url + "/user/" + user_id + "/policyMapping",
            body=json.dumps(dict(policies=policies)),
            entity_name="user",
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def get_user(self, user_email: str):
        """
        Get user's account mapping.

        # Returns
        (Object): Spotinst API response
        """
        query_params = dict(userEmail=user_email)

        response = self.send_get(
            url=self.__base_setup_url + "/accountUserMapping", entity_name="user",
            query_params=query_params
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"]

    def create_programmatic_user(self, name: str, description: str, accounts=None, policies=None):
        """
        Create a programmatic user

        # Arguments
        name (String): Name of the programmatic user
        description (String): Brief description of the user
        accounts (List): All the accounts the programmatic user will have access to
        policies (List): All the policies the programmatic user will have access to

        # Returns
        (Object): Spotinst API response
        """
        if accounts is None and policies is None:
            raise ValueError(
                "Either accounts or policies must be provided in create_programmatic_user"
            )

        payload = {
            "description": description,
            "name": name,
        }
        if accounts is not None:
            payload["accounts"] = accounts
        if policies is not None:
            payload["policies"] = policies

        response = self.send_post(
            url=self.__base_setup_url + "/user/programmatic",
            body=json.dumps(payload),
            entity_name="programmaticUser",
        )
        formatted_response = self.convert_json(
            response, self.camel_to_underscore)
        return formatted_response["response"]["items"][0]

    def get_access_policy_actions(self, category: str = None, name: str = None, resource_pattern: str = None, scope: str = None, service: str = None):
        """
        Get actions for access policies.

        # Returns
        (Object): Spotinst API response
        """
        query_params = dict(category=category, name=name,
                            resourcePattern=resource_pattern, scope=scope, service=service)

        response = self.send_get(
            url=self.__base_setup_url + "/access/policyAction", entity_name="user",
            query_params=query_params
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"]

    def create_access_policy(self, policy: admin_org.AccessPolicy):
        """
        Create an access policy

        # Arguments
        policy (AccessPolicy): AccessPolicy Object

        # Returns
        (Object): Spotinst API response 
        """
        request = admin_org.AccessPolicyCreationRequest(policy)

        excluded_policy_dict = self.exclude_missing(
            json.loads(request.toJSON()))

        formatted_policy_dict = self.convert_json(
            excluded_policy_dict, self.underscore_to_camel)

        body_json = json.dumps(formatted_policy_dict)

        policy_response = self.send_post(
            body=body_json,
            url=self.__base_setup_url + "/access/policy", entity_name="policy")

        formatted_response = self.convert_json(
            policy_response, self.camel_to_underscore)

        ret_val = formatted_response["response"]["items"][0]

        return ret_val

    def update_access_policy(self, policy_id: str, policy: admin_org.AccessPolicy):
        """
        Updates an access policy settings.

        # Arguments
        policy_id (String): Policy ID
        policy (AccessPolicy): AccessPolicy Object

        # Returns
        (Object): Spotinst API response
        """
        request = admin_org.AccessPolicyUpdationRequest(policy)

        excluded_policy_dict = self.exclude_missing(
            json.loads(request.toJSON()))

        formatted_policy_dict = self.convert_json(
            excluded_policy_dict, self.underscore_to_camel)

        body_json = json.dumps(formatted_policy_dict)

        policy_response = self.send_put(
            body=body_json,
            url=self.__base_setup_url + "/access/policy/" + policy_id, entity_name="policy")

        formatted_response = self.convert_json(
            policy_response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def delete_access_policy(self, policy_id: str):
        """
        Deletes an access policy.

        # Arguments
        policy_id (String): Policy ID

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_delete(
            url=self.__base_setup_url + "/access/policy/" + policy_id, entity_name="policy"
        )

        return response

    def get_user_groups(self):
        """
        Retrieves all user-groups from an organization.

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_get(
            url=self.__base_setup_url + "/access/userGroup", entity_name="usergroup"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"]

    def create_user_group(self, user_group: admin_org.UserGroup):
        """
        Create a new User Group

        # Arguments
        group (UserGroup): UserGroup Object

        # Returns
        (Object): Spotinst API response 
        """
        request = admin_org.UserGroupCreationRequest(user_group)

        excluded_group_dict = self.exclude_missing(
            json.loads(request.toJSON()))

        formatted_group_dict = self.convert_json(
            excluded_group_dict, self.underscore_to_camel)

        body_json = json.dumps(formatted_group_dict)

        policy_response = self.send_post(
            body=body_json,
            url=self.__base_setup_url + "/access/userGroup", entity_name="usergroup")

        formatted_response = self.convert_json(
            policy_response, self.camel_to_underscore)

        ret_val = formatted_response["response"]["items"][0]

        return ret_val

    def get_user_group_details(self, user_group_id: str):
        """
        Get the details of a user Group

        # Arguments
        user_group_id (String): User Group ID

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_get(
            url=self.__base_setup_url + "/access/userGroup/" + user_group_id, entity_name="usergroup"
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["items"][0]

    def delete_user_group(self, user_group_id: str):
        """
        Delete a user group.

        # Arguments
        user_group_id (String): User Group ID

        # Returns
        (Object): Spotinst API response
        """
        return self.send_delete(
            url=self.__base_setup_url + "/access/userGroup/" + user_group_id, entity_name="usergroup"
        )

    def update_user_group_to_user_mapping(self, user_group_id: str, user_ids: List[str]):
        """
        Update the mapping of a given user group to users

        # Arguments
        user_group_id (String): Identifier of a usergroup.
        user_ids (List): The users to register under the given user group (should be existing users only)

        # Returns
        (Object): Spotinst API response
        """
        response = self.send_put(
            url=self.__base_setup_url + "/access/userGroup/" + user_group_id + "/userMapping",
            body=json.dumps(dict(userIds=user_ids)),
            entity_name="usergroup",
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]

    def update_user_group_to_policy_mapping(self, user_group_id: str, policies):
        """
        Update the mapping of a given user group to policies

        # Arguments
        user_group_id (String): Identifier of a user group.
        policies (List): The policies to register under the given user group (should be existing policies only)

        # Returns
        (Object): Spotinst API response
        """

        response = self.send_put(
            url=self.__base_setup_url + "/access/userGroup/" +
            user_group_id + "/policyMapping",
            body=json.dumps(dict(policies=policies)),
            entity_name="usergroup",
        )

        formatted_response = self.convert_json(
            response, self.camel_to_underscore)

        return formatted_response["response"]["status"]
